/* cvbsctl-sunxi.c 
 *
 * (c) 1.6.2014 Ingo D. Rullhusen
 *
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/spinlock.h>

#include <linux/device.h>
#include <linux/kdev_t.h>

#include <asm/io.h>



#define TVE0_BASE 0x01c0a000
#define TVE1_BASE 0x01c1b000


#define TVE_004_REG_OFFSET 0x004

#define COLORBARMODE_MASK 0x00000100
#define COLORBARTYPE_MASK 0x00000200

#define TVMODE_MASK 0x0000000f
#define TVMODE_SHIFT 0


#define TVE_014_REG_OFFSET 0x014

#define FRONT_PORCH_MASK 0x00000fff
#define FRONT_PORCH_SHIFT 0

#define BACK_PORCH_MASK 0x01ff0000
#define BACK_PORCH_SHIFT 16


#define TVE_01C_REG_OFFSET 0x01c

#define FIRST_VIDEO_LINE_MASK 0x00ff0000
#define FIRST_VIDEO_LINE_SHIFT 16

#define NUM_LINES_MASK 0x000007ff
#define NUM_LINES_SHIFT 0


#define TVE_114_REG_OFFSET 0x114


#define TVE_124_REG_OFFSET 0x124

#define ACTIVE_LINE_MASK 0x00000fff
#define ACTIVE_LINE_SHIFT 0


#define TVE_130_REG_OFFSET 0x130

#define RESYNC_FIELD_MASK 0x80000000
#define RESYNC_FIELD_SHIFT 31

#define RESYNC_DISABLE_MASK 0x40000000
#define RESYNC_DISABLE_SHIFT 30

#define RESYNC_LINE_NUM_MASK 0x07ff0000
#define RESYNC_LINE_NUM_SHIFT 16

#define RESYNC_PIXEL_NUM_MASK 0x000007ff
#define RESYNC_PIXEL_NUM_SHIFT 0


#define TVE_13C_REG_OFFSET 0x13c
#define TVE_00C_REG_OFFSET 0x00c


#define TVE_020_REG_OFFSET 0x020

#define BLANK_LEVEL_MASK 0x03ff0000
#define BLANK_LEVEL_SHIFT 16

#define BLACK_LEVEL_MASK 0x000003ff
#define BLACK_LEVEL_SHIFT 0


#define TVE_10C_REG_OFFSET 0x10c


#define TVE_128_REG_OFFSET 0x128

#define CHROMA_BW_MASK 0x00030000
#define CHROMA_BW_SHIFT 16


#define TVE_118_REG_OFFSET 0x118
#define TVE_12C_REG_OFFSET 0x12c


#define TVE_008_REG_OFFSET 0x008

#define LOW_BIAS_MASK 0xf0000000
#define LOW_BIAS_SHIFT 28

#define TERMINAL_MODE_MASK 0x00030000
#define TERMINAL_MODE_SHIFT 16


#define TVE_024_REG_OFFSET 0x024

#define DAC3_AMPL_MASK 0x1f000000
#define DAC3_AMPL_SHIFT 24

#define DAC2_AMPL_MASK 0x001f0000
#define DAC2_AMPL_SHIFT 16

#define DAC1_AMPL_MASK 0x00001f00
#define DAC1_AMPL_SHIFT 8

#define DAC0_AMPL_MASK 0x0000001f
#define DAC0_AMPL_SHIFT 0


#define TVE_110_REG_OFFSET 0x110

#define TINT_MASK 0x00ff0000
#define TINT_SHIFT 16


#define TVE_120_REG_OFFSET 0x120

#define WHITE_LEVEL_MASK 0x000003ff
#define WHITE_LEVEL_SHIFT 0




static DEFINE_SPINLOCK(slock);



typedef struct
{
   unsigned long *ptr_004;
   unsigned long *ptr_014;
   unsigned long *ptr_01c;
   unsigned long *ptr_114;
   unsigned long *ptr_124;
   unsigned long *ptr_130;
   unsigned long *ptr_13c;
   unsigned long *ptr_00c;
   unsigned long *ptr_020;
   unsigned long *ptr_10c;
   unsigned long *ptr_128;
   unsigned long *ptr_118;
   unsigned long *ptr_12c;
   unsigned long *ptr_008;
   unsigned long *ptr_024;
   unsigned long *ptr_110;
   unsigned long *ptr_120;

   unsigned long buff_020, buff_024, buff_130;

} tve_data;

static tve_data tve0_data;
static tve_data tve1_data;

static void tve_data_init( struct device *tve )
{
   tve_data *data = dev_get_drvdata(tve);

/*** PAL init ***/
   writel( 0x07030001, data->ptr_004 );
   writel( 0x008a0018, data->ptr_014 );
   writel( 0x00160271, data->ptr_01c );
   writel( 0x0016447e, data->ptr_114 );
   writel( 0x000005a0, data->ptr_124 );
   writel( 0x800D000C, data->ptr_130 );
   writel( 0x00000000, data->ptr_13c );
   writel( 0x00000120, data->ptr_00c );
   writel( 0x00fc00fc, data->ptr_020 );
   writel( 0x00002828, data->ptr_10c );
   writel( 0x00000002, data->ptr_128 );
   writel( 0x0000e0e0, data->ptr_118 );
   writel( 0x00000101, data->ptr_12c );
   writel( 0x413f0008, data->ptr_008 );
   writel( 0x18181818, data->ptr_024 );
   writel( 0x00000000, data->ptr_110 );
   writel( 0x00000000, data->ptr_120 );
}


/*   /sys/class/colorbar-sunxi entries for tve  */

static ssize_t colorbartype_show(struct device *dev,struct device_attribute *attr, char *buf)
{
   ssize_t size;
   unsigned long value;
   tve_data *data = dev_get_drvdata(dev);

   printk( KERN_INFO "show colorbartype\n" );

   value = readl( data->ptr_004 );
      
   size = scnprintf( buf, PAGE_SIZE, "%d", (value&COLORBARTYPE_MASK)>0 );

   return(size);
}

static ssize_t colorbartype_store(struct device *dev,struct device_attribute *attr, const char *buf, size_t size)
{
   unsigned long value;
   int colorbartype = 0;
   tve_data *data = dev_get_drvdata(dev);

   printk( KERN_INFO "store colorbartype\n" );

   sscanf(buf,"%d",&colorbartype);

   spin_lock(&slock);

   value = readl( data->ptr_004 );

   if ( colorbartype ) { value |=  COLORBARTYPE_MASK; }
   else        { value &= ~COLORBARTYPE_MASK; }

   writel( value, data->ptr_004 );

   spin_unlock(&slock);

   return(size);
}

static DEVICE_ATTR(colorbartype, 0644, colorbartype_show, colorbartype_store);

static ssize_t colorbar_show(struct device *dev,struct device_attribute *attr, char *buf)
{
   ssize_t size;
   unsigned long value;
   tve_data *data = dev_get_drvdata(dev);

   printk( KERN_INFO "show colorbar\n" );

   value = readl( data->ptr_004 );
      
   size = scnprintf( buf, PAGE_SIZE, "%d", (value&COLORBARMODE_MASK)>0 );

   return(size);
}

static ssize_t colorbar_store(struct device *dev,struct device_attribute *attr, const char *buf, size_t size)
{
   unsigned long value;
   int enable = 0;
   tve_data *data = dev_get_drvdata(dev);

   printk( KERN_INFO "store colorbar\n" );

   sscanf(buf,"%d",&enable);

   spin_lock(&slock);

   value = readl( data->ptr_004 );

   if ( ! (value&COLORBARMODE_MASK) )
   {
      data->buff_020 = readl( data->ptr_020 );
      data->buff_024 = readl( data->ptr_024 );
      data->buff_130 = readl( data->ptr_130 );
   }

   if ( enable )
   {
      value |=  COLORBARMODE_MASK;
      writel( value, data->ptr_004 );
   }
   else
   {
      value &= ~COLORBARMODE_MASK;
      writel( value, data->ptr_004 );
      writel( data->buff_020, data->ptr_020 );
      writel( data->buff_024, data->ptr_024 );
      writel( data->buff_130, data->ptr_130 );
   }

   spin_unlock(&slock);

   return(size);
}

static DEVICE_ATTR(colorbar, 0644, colorbar_show, colorbar_store);

static ssize_t tvmode_show(struct device *dev,struct device_attribute *attr, char *buf)
{
   ssize_t size;
   unsigned long value;
   tve_data *data = dev_get_drvdata(dev);

   printk( KERN_INFO "show tvmode\n" );

   value = readl( data->ptr_004 );
      
   size = scnprintf( buf, PAGE_SIZE, "%lu", (value&TVMODE_MASK)>>TVMODE_SHIFT );

   return(size);
}

static ssize_t tvmode_store(struct device *dev,struct device_attribute *attr, const char *buf, size_t size)
{
   unsigned long value;
   unsigned long tvmode = 0;
   tve_data *data = dev_get_drvdata(dev);

   printk( KERN_INFO "store tvmode\n" );

   sscanf(buf,"%lu",&tvmode);
   tvmode = (tvmode<<TVMODE_SHIFT)&TVMODE_MASK;

   spin_lock(&slock);

   value = readl( data->ptr_004 );
   value &= ~TVMODE_MASK;
   value |= tvmode;

   writel( value, data->ptr_004 );

   spin_unlock(&slock);

   return(size);
}

static DEVICE_ATTR(tvmode, 0644, tvmode_show, tvmode_store);

static ssize_t frontporch_show(struct device *dev,struct device_attribute *attr, char *buf)
{
   ssize_t size;
   unsigned long value;
   tve_data *data = dev_get_drvdata(dev);

   printk( KERN_INFO "show front porch\n" );

   value = readl( data->ptr_014 );
      
   size = scnprintf( buf, PAGE_SIZE, "%lu", (value&FRONT_PORCH_MASK)>>FRONT_PORCH_SHIFT );

   return(size);
}

static ssize_t frontporch_store(struct device *dev,struct device_attribute *attr, const char *buf, size_t size)
{
   unsigned long value;
   unsigned long frontporch = 0;
   tve_data *data = dev_get_drvdata(dev);

   printk( KERN_INFO "store front porch\n" );

   sscanf(buf,"%lu",&frontporch);
   frontporch = (frontporch<<FRONT_PORCH_SHIFT)&FRONT_PORCH_MASK;

   spin_lock(&slock);

   value = readl( data->ptr_014 );
   value &= ~FRONT_PORCH_MASK;
   value |= frontporch;

   writel( value, data->ptr_014 );

   spin_unlock(&slock);

   return(size);
}

static DEVICE_ATTR(frontporch, 0644, frontporch_show, frontporch_store);

static ssize_t backporch_show(struct device *dev,struct device_attribute *attr, char *buf)
{
   ssize_t size;
   unsigned long value;
   tve_data *data = dev_get_drvdata(dev);

   printk( KERN_INFO "show front porch\n" );

   value = readl( data->ptr_014 );
      
   size = scnprintf( buf, PAGE_SIZE, "%lu", (value&BACK_PORCH_MASK)>>BACK_PORCH_SHIFT );

   return(size);
}

static ssize_t backporch_store(struct device *dev,struct device_attribute *attr, const char *buf, size_t size)
{
   unsigned long value;
   unsigned long backporch = 0;
   tve_data *data = dev_get_drvdata(dev);

   printk( KERN_INFO "store front porch\n" );

   sscanf(buf,"%lu",&backporch);
   backporch = (backporch<<BACK_PORCH_SHIFT)&BACK_PORCH_MASK;

   spin_lock(&slock);

   value = readl( data->ptr_014 );
   value &= ~BACK_PORCH_MASK;
   value |= backporch;

   writel( value, data->ptr_014 );

   spin_unlock(&slock);

   return(size);
}

static DEVICE_ATTR(backporch, 0644, backporch_show, backporch_store);

static ssize_t firstline_show(struct device *dev,struct device_attribute *attr, char *buf)
{
   ssize_t size;
   unsigned long value;
   tve_data *data = dev_get_drvdata(dev);

   printk( KERN_INFO "show first line\n" );

   value = readl( data->ptr_01c );
      
   size = scnprintf( buf, PAGE_SIZE, "%lu", (value&FIRST_VIDEO_LINE_MASK)>>FIRST_VIDEO_LINE_SHIFT );

   return(size);
}

static ssize_t firstline_store(struct device *dev,struct device_attribute *attr, const char *buf, size_t size)
{
   unsigned long value;
   unsigned long firstline = 0;
   tve_data *data = dev_get_drvdata(dev);

   printk( KERN_INFO "store first line\n" );

   sscanf(buf,"%lu",&firstline);

   spin_lock(&slock);

   value = readl( data->ptr_01c );
   value &= ~FIRST_VIDEO_LINE_MASK;
   value |= (firstline<<FIRST_VIDEO_LINE_SHIFT)&FIRST_VIDEO_LINE_MASK;

   writel( value, data->ptr_01c );

   spin_unlock(&slock);

   return(size);
}

static DEVICE_ATTR(firstline, 0644, firstline_show, firstline_store);

static ssize_t numlines_show(struct device *dev,struct device_attribute *attr, char *buf)
{
   ssize_t size;
   unsigned long value;
   tve_data *data = dev_get_drvdata(dev);

   printk( KERN_INFO "show numlines\n" );

   value = readl( data->ptr_01c );
      
   size = scnprintf( buf, PAGE_SIZE, "%lu", (value&NUM_LINES_MASK)>>NUM_LINES_SHIFT );

   return(size);
}

static ssize_t numlines_store(struct device *dev,struct device_attribute *attr, const char *buf, size_t size)
{
   unsigned long value;
   unsigned long numlines = 0;
   tve_data *data = dev_get_drvdata(dev);

   printk( KERN_INFO "store numlines\n" );

   sscanf(buf,"%lu",&numlines);
   numlines = (numlines<<NUM_LINES_SHIFT)&NUM_LINES_MASK;

   spin_lock(&slock);

   value = readl( data->ptr_01c );
   value &= ~NUM_LINES_MASK;
   value |= numlines;

   writel( value, data->ptr_01c );

   spin_unlock(&slock);

   return(size);
}

static DEVICE_ATTR(numlines, 0644, numlines_show, numlines_store);

static ssize_t lowbias_show(struct device *dev,struct device_attribute *attr, char *buf)
{
   ssize_t size;
   unsigned long value, lowbias;
   tve_data *data = dev_get_drvdata(dev);

   value = readl( data->ptr_008 );
   lowbias = (value&LOW_BIAS_MASK)>>LOW_BIAS_SHIFT;

   size = scnprintf( buf, PAGE_SIZE, "%lu", lowbias );

   printk( KERN_INFO "low bias is %lu µA\n", lowbias*250 + 250 );

   return(size);
}

static ssize_t lowbias_store(struct device *dev,struct device_attribute *attr, const char *buf, size_t size)
{
   unsigned long value;
   unsigned long lowbias = 0;
   tve_data *data = dev_get_drvdata(dev);

   sscanf(buf,"%lu",&lowbias);
   lowbias = (lowbias<<LOW_BIAS_SHIFT)&LOW_BIAS_MASK;

   spin_lock(&slock);

   value = readl( data->ptr_008 );
   value &= ~LOW_BIAS_MASK;
   value |= lowbias;

   writel( value, data->ptr_008 );

   spin_unlock(&slock);

   return(size);
}

static DEVICE_ATTR(lowbias, 0644, lowbias_show, lowbias_store);

static ssize_t terminalmode_show(struct device *dev,struct device_attribute *attr, char *buf)
{
   ssize_t size;
   unsigned long value, mode;
   tve_data *data = dev_get_drvdata(dev);

   value = readl( data->ptr_008 );
   mode = (value&TERMINAL_MODE_MASK)>>TERMINAL_MODE_SHIFT;

   size = scnprintf( buf, PAGE_SIZE, "%lu", mode );

   switch ( mode )
   { case 0: printk( KERN_INFO "terminal mode is 175 Ohm\n" ); break;
     case 2: printk( KERN_INFO "terminal mode is 75 Ohm\n" ); break;
     case 3: printk( KERN_INFO "terminal mode is 37.5 Ohm\n" ); break;
     default: printk( KERN_INFO "terminal mode is not specified\n" ); break;
   }

   return(size);
}

static ssize_t terminalmode_store(struct device *dev,struct device_attribute *attr, const char *buf, size_t size)
{
   unsigned long value;
   unsigned long mode = 0;
   tve_data *data = dev_get_drvdata(dev);

   sscanf(buf,"%lu",&mode);

   spin_lock(&slock);

   value = readl( data->ptr_008 );
   value &= ~TERMINAL_MODE_MASK;
   value |= (mode<<TERMINAL_MODE_SHIFT)&TERMINAL_MODE_MASK;

   writel( value, data->ptr_008 );

   spin_unlock(&slock);

   return(size);
}

static DEVICE_ATTR(terminalmode, 0644, terminalmode_show, terminalmode_store);

static ssize_t blanklevel_show(struct device *dev,struct device_attribute *attr, char *buf)
{
   ssize_t size;
   unsigned long value;
   tve_data *data = dev_get_drvdata(dev);

   printk( KERN_INFO "show blaNk level\n" );

   value = readl( data->ptr_020 );
      
   size = scnprintf( buf, PAGE_SIZE, "%lu", (value&BLANK_LEVEL_MASK)>>BLANK_LEVEL_SHIFT );

   return(size);
}

static ssize_t blanklevel_store(struct device *dev,struct device_attribute *attr, const char *buf, size_t size)
{
   unsigned long value;
   unsigned long blanklevel;
   tve_data *data = dev_get_drvdata(dev);

   printk( KERN_INFO "store blaNk level\n" );

   sscanf(buf,"%lu",&blanklevel);

   spin_lock(&slock);

   value = readl( data->ptr_020 );
   value &= ~BLANK_LEVEL_MASK;
   value |= (blanklevel<<BLANK_LEVEL_SHIFT)&BLANK_LEVEL_MASK;

   writel( value, data->ptr_020 );

   spin_unlock(&slock);

   return(size);
}

static DEVICE_ATTR(blanklevel, 0644, blanklevel_show, blanklevel_store);

static ssize_t blacklevel_show(struct device *dev,struct device_attribute *attr, char *buf)
{
   ssize_t size;
   unsigned long value;
   tve_data *data = dev_get_drvdata(dev);

   printk( KERN_INFO "show blaCk level\n" );

   value = readl( data->ptr_020 );
      
   size = scnprintf( buf, PAGE_SIZE, "%lu", (value&BLACK_LEVEL_MASK)>>BLACK_LEVEL_SHIFT );

   return(size);
}

static ssize_t blacklevel_store(struct device *dev,struct device_attribute *attr, const char *buf, size_t size)
{
   unsigned long value;
   unsigned long blacklevel;
   tve_data *data = dev_get_drvdata(dev);

   printk( KERN_INFO "store blaCk level\n" );

   sscanf(buf,"%lu",&blacklevel);

   spin_lock(&slock);

   value = readl( data->ptr_020 );
   value &= ~BLACK_LEVEL_MASK;
   value |= (blacklevel<<BLACK_LEVEL_SHIFT)&BLACK_LEVEL_MASK;

   writel( value, data->ptr_020 );

   spin_unlock(&slock);

   return(size);
}

static DEVICE_ATTR(blacklevel, 0644, blacklevel_show, blacklevel_store);

static ssize_t tint_show(struct device *dev,struct device_attribute *attr, char *buf)
{
   ssize_t size;
   unsigned long value;
   tve_data *data = dev_get_drvdata(dev);

   printk( KERN_INFO "show tint\n" );

   value = readl( data->ptr_110 );
      
   size = scnprintf( buf, PAGE_SIZE, "%lu", (value&TINT_MASK)>>TINT_SHIFT );

   return(size);
}

static ssize_t tint_store(struct device *dev,struct device_attribute *attr, const char *buf, size_t size)
{
   unsigned long value;
   unsigned long tint;
   tve_data *data = dev_get_drvdata(dev);

   printk( KERN_INFO "store tint\n" );

   sscanf(buf,"%lu",&tint);

   spin_lock(&slock);

   value = readl( data->ptr_110 );
   value &= ~TINT_MASK;
   value |= (tint<<TINT_SHIFT)&TINT_MASK;

   writel( value, data->ptr_110 );

   spin_unlock(&slock);

   return(size);
}

static DEVICE_ATTR(tint, 0644, tint_show, tint_store);

static ssize_t whitelevel_show(struct device *dev,struct device_attribute *attr, char *buf)
{
   ssize_t size;
   unsigned long value;
   tve_data *data = dev_get_drvdata(dev);

   printk( KERN_INFO "show white level\n" );

   value = readl( data->ptr_120 );
      
   size = scnprintf( buf, PAGE_SIZE, "%lu", (value&WHITE_LEVEL_MASK)>>WHITE_LEVEL_SHIFT );

   return(size);
}

static ssize_t whitelevel_store(struct device *dev,struct device_attribute *attr, const char *buf, size_t size)
{
   unsigned long value;
   unsigned long whitelevel;
   tve_data *data = dev_get_drvdata(dev);

   printk( KERN_INFO "store white level\n" );

   sscanf(buf,"%lu",&whitelevel);

   spin_lock(&slock);

   value = readl( data->ptr_120 );
   value &= ~WHITE_LEVEL_MASK;
   value |= (whitelevel<<WHITE_LEVEL_SHIFT)&WHITE_LEVEL_MASK;

   writel( value, data->ptr_120 );

   spin_unlock(&slock);

   return(size);
}

static DEVICE_ATTR(whitelevel, 0644, whitelevel_show, whitelevel_store);

static ssize_t activeline_show(struct device *dev,struct device_attribute *attr, char *buf)
{
   ssize_t size;
   unsigned long value;
   tve_data *data = dev_get_drvdata(dev);

   printk( KERN_INFO "show active line\n" );

   value = readl( data->ptr_124 );
      
   size = scnprintf( buf, PAGE_SIZE, "%lu", (value&ACTIVE_LINE_MASK)>>ACTIVE_LINE_SHIFT );

   return(size);
}

static ssize_t activeline_store(struct device *dev,struct device_attribute *attr, const char *buf, size_t size)
{
   unsigned long value;
   unsigned long activeline;
   tve_data *data = dev_get_drvdata(dev);

   printk( KERN_INFO "store active line\n" );

   sscanf(buf,"%lu",&activeline);

   spin_lock(&slock);

   value = readl( data->ptr_124 );
   value &= ~ACTIVE_LINE_MASK;
   value |= (activeline<<ACTIVE_LINE_SHIFT)&ACTIVE_LINE_MASK;

   writel( value, data->ptr_124 );

   spin_unlock(&slock);

   return(size);
}

static DEVICE_ATTR(activeline, 0644, activeline_show, activeline_store);

static ssize_t chromaBW_show(struct device *dev,struct device_attribute *attr, char *buf)
{
   ssize_t size;
   unsigned long value, cbw;
   tve_data *data = dev_get_drvdata(dev);

   value = readl( data->ptr_128 );
   cbw = (value&CHROMA_BW_MASK)>>CHROMA_BW_SHIFT;

   size = scnprintf( buf, PAGE_SIZE, "%lu", cbw );

   switch ( cbw )
   { case 0: printk( KERN_INFO "chroma bandwidth is 0.7 MHz\n" ); break;
     case 1: printk( KERN_INFO "chroma bandwidth is 1.2 MHz\n" ); break;
     case 2: printk( KERN_INFO "chroma bandwidth is 1.8 MHz\n" ); break;
     case 3: printk( KERN_INFO "chroma bandwidth is 2.5 MHz\n" ); break;
     default: printk( KERN_INFO "chroma bandwidth is not specified\n" ); break;
   }

   return(size);
}

static ssize_t chromaBW_store(struct device *dev,struct device_attribute *attr, const char *buf, size_t size)
{
   unsigned long value;
   unsigned long cbw;
   tve_data *data = dev_get_drvdata(dev);

   printk( KERN_INFO "store chroma bandwidth\n" );

   sscanf(buf,"%lu",&cbw);

   spin_lock(&slock);

   value = readl( data->ptr_128 );
   value &= ~CHROMA_BW_MASK;
   value |= (cbw<<CHROMA_BW_SHIFT)&CHROMA_BW_MASK;

   writel( value, data->ptr_128 );

   spin_unlock(&slock);

   return(size);
}

static DEVICE_ATTR(chroma_bw, 0644, chromaBW_show, chromaBW_store);

static ssize_t rsfield_show(struct device *dev,struct device_attribute *attr, char *buf)
{
   ssize_t size;
   unsigned long value;
   tve_data *data = dev_get_drvdata(dev);

   printk( KERN_INFO "show resync field\n" );

   value = readl( data->ptr_130 );
      
   size = scnprintf( buf, PAGE_SIZE, "%d", (value&RESYNC_FIELD_MASK)>0 );

   return(size);
}

static ssize_t rsfield_store(struct device *dev,struct device_attribute *attr, const char *buf, size_t size)
{
   unsigned long value;
   int enable = 0;
   tve_data *data = dev_get_drvdata(dev);

   printk( KERN_INFO "store resync field\n" );

   sscanf(buf,"%d",&enable);

   spin_lock(&slock);

   value = readl( data->ptr_130 );

   if ( enable ) { value |=  RESYNC_FIELD_MASK; }
   else          { value &= ~RESYNC_FIELD_MASK; }

   writel( value, data->ptr_130 );

   spin_unlock(&slock);

   return(size);
}

static DEVICE_ATTR(resync_field, 0644, rsfield_show, rsfield_store);

static ssize_t rsdisable_show(struct device *dev,struct device_attribute *attr, char *buf)
{
   ssize_t size;
   unsigned long value;
   tve_data *data = dev_get_drvdata(dev);

   printk( KERN_INFO "show resync disable\n" );

   value = readl( data->ptr_130 );
      
   size = scnprintf( buf, PAGE_SIZE, "%d", (value&RESYNC_DISABLE_MASK)>0 );

   return(size);
}

static ssize_t rsdisable_store(struct device *dev,struct device_attribute *attr, const char *buf, size_t size)
{
   unsigned long value;
   int enable = 0;
   tve_data *data = dev_get_drvdata(dev);

   printk( KERN_INFO "store resync disable\n" );

   sscanf(buf,"%d",&enable);

   spin_lock(&slock);

   value = readl( data->ptr_130 );

   if ( enable ) { value |=  RESYNC_DISABLE_MASK; }
   else          { value &= ~RESYNC_DISABLE_MASK; }

   writel( value, data->ptr_130 );

   spin_unlock(&slock);

   return(size);
}

static DEVICE_ATTR(resync_disable, 0644, rsdisable_show, rsdisable_store);

static ssize_t rslinenum_show(struct device *dev,struct device_attribute *attr, char *buf)
{
   ssize_t size;
   unsigned long value;
   tve_data *data = dev_get_drvdata(dev);

   printk( KERN_INFO "show resync line num\n" );

   value = readl( data->ptr_130 );
      
   size = scnprintf( buf, PAGE_SIZE, "%lu", (value&RESYNC_LINE_NUM_MASK)>>RESYNC_LINE_NUM_SHIFT );

   return(size);
}

static ssize_t rslinenum_store(struct device *dev,struct device_attribute *attr, const char *buf, size_t size)
{
   unsigned long value;
   unsigned long linenum;
   tve_data *data = dev_get_drvdata(dev);

   printk( KERN_INFO "store resync linenum\n" );

   sscanf(buf,"%lu",&linenum);

   spin_lock(&slock);

   value = readl( data->ptr_130 );
   value &= ~RESYNC_LINE_NUM_MASK;
   value |= (linenum<<RESYNC_LINE_NUM_SHIFT)&RESYNC_LINE_NUM_MASK;

   writel( value, data->ptr_130 );

   spin_unlock(&slock);

   return(size);
}

static DEVICE_ATTR(resync_linenum, 0644, rslinenum_show, rslinenum_store);


static ssize_t rspixelnum_show(struct device *dev,struct device_attribute *attr, char *buf)
{
   ssize_t size;
   unsigned long value;
   tve_data *data = dev_get_drvdata(dev);

   printk( KERN_INFO "show resync pixel num\n" );

   value = readl( data->ptr_130 );
      
   size = scnprintf( buf, PAGE_SIZE, "%lu", (value&RESYNC_PIXEL_NUM_MASK)>>RESYNC_PIXEL_NUM_SHIFT );

   return(size);
}

static ssize_t rspixelnum_store(struct device *dev,struct device_attribute *attr, const char *buf, size_t size)
{
   unsigned long value;
   unsigned long pixelnum;
   tve_data *data = dev_get_drvdata(dev);

   printk( KERN_INFO "store resync pixelnum\n" );

   sscanf(buf,"%lu",&pixelnum);

   spin_lock(&slock);

   value = readl( data->ptr_130 );
   value &= ~RESYNC_PIXEL_NUM_MASK;
   value |= (pixelnum<<RESYNC_PIXEL_NUM_SHIFT)&RESYNC_PIXEL_NUM_MASK;

   writel( value, data->ptr_130 );

   spin_unlock(&slock);

   return(size);
}

static DEVICE_ATTR(resync_pixelnum, 0644, rspixelnum_show, rspixelnum_store);

static ssize_t dacampl_show(struct device *dev,struct device_attribute *attr, char *buf)
{
   ssize_t size;
   unsigned long value;
   tve_data *data = dev_get_drvdata(dev);

   printk( KERN_INFO "show DAC 3 2 1 0 amplitudes\n" );

   value = readl( data->ptr_024 );
      
   size = scnprintf( buf, PAGE_SIZE, "%lu %lu %lu %lu",
                     (value&DAC3_AMPL_MASK)>>DAC3_AMPL_SHIFT,
                     (value&DAC2_AMPL_MASK)>>DAC2_AMPL_SHIFT,
                     (value&DAC1_AMPL_MASK)>>DAC1_AMPL_SHIFT,
                     (value&DAC0_AMPL_MASK)>>DAC0_AMPL_SHIFT  );

   return(size);
}

static ssize_t dacampl_store(struct device *dev,struct device_attribute *attr, const char *buf, size_t size)
{
   unsigned long value;
   unsigned long DACampl[4];
   tve_data *data = dev_get_drvdata(dev);

   printk( KERN_INFO "store DAC 3 2 1 0 amplitudes\n" );

   sscanf(buf,"%lu %lu %lu %lu",&DACampl[3],&DACampl[2],&DACampl[1],&DACampl[0]);

   spin_lock(&slock);

   value = readl( data->ptr_024 );

   value &= ~DAC3_AMPL_MASK;
   value |= (DACampl[3]<<DAC3_AMPL_SHIFT)&DAC3_AMPL_MASK;

   value &= ~DAC2_AMPL_MASK;
   value |= (DACampl[2]<<DAC2_AMPL_SHIFT)&DAC2_AMPL_MASK;

   value &= ~DAC1_AMPL_MASK;
   value |= (DACampl[1]<<DAC1_AMPL_SHIFT)&DAC1_AMPL_MASK;

   value &= ~DAC0_AMPL_MASK;
   value |= (DACampl[0]<<DAC0_AMPL_SHIFT)&DAC0_AMPL_MASK;

   writel( value, data->ptr_024 );

   spin_unlock(&slock);

   return(size);
}

static DEVICE_ATTR(dac_ampl, 0644, dacampl_show, dacampl_store);




static const struct attribute *tve_attrs[] =
{
   &dev_attr_colorbartype.attr,
   &dev_attr_colorbar.attr,
   &dev_attr_tvmode.attr,
   &dev_attr_firstline.attr,
   &dev_attr_numlines.attr,
   &dev_attr_frontporch.attr,
   &dev_attr_backporch.attr,
   &dev_attr_lowbias.attr,
   &dev_attr_terminalmode.attr,
   &dev_attr_blanklevel.attr,
   &dev_attr_blacklevel.attr,
   &dev_attr_tint.attr,
   &dev_attr_whitelevel.attr,
   &dev_attr_activeline.attr,
   &dev_attr_chroma_bw.attr,
   &dev_attr_resync_field.attr,
   &dev_attr_resync_disable.attr,
   &dev_attr_resync_linenum.attr,
   &dev_attr_resync_pixelnum.attr,
   &dev_attr_dac_ampl.attr,
   NULL,
};

static const struct attribute_group tve_attr_group =
{
   .attrs = (struct attribute **) tve_attrs
};

static struct device *tve0;
static struct device *tve1;




static struct class_attribute cvbsctl_class_attrs[] =
{ __ATTR_NULL };

static struct class cvbsctl_class =
{
   .name =         "cvbsctl-sunxi",
   .owner =        THIS_MODULE,
   .class_attrs =  cvbsctl_class_attrs,
};



static int __init sunxi_cvbsctl_init(void)
{
   int err = 0;
   printk( KERN_INFO "init cvbsctl\n" );

   tve0_data.ptr_004 = ioremap( TVE0_BASE+TVE_004_REG_OFFSET, 0x004 );
   tve0_data.ptr_014 = ioremap( TVE0_BASE+TVE_014_REG_OFFSET, 0x004 );
   tve0_data.ptr_01c = ioremap( TVE0_BASE+TVE_01C_REG_OFFSET, 0x004 );
   tve0_data.ptr_114 = ioremap( TVE0_BASE+TVE_114_REG_OFFSET, 0x004 );
   tve0_data.ptr_124 = ioremap( TVE0_BASE+TVE_124_REG_OFFSET, 0x004 );
   tve0_data.ptr_130 = ioremap( TVE0_BASE+TVE_130_REG_OFFSET, 0x004 );
   tve0_data.ptr_13c = ioremap( TVE0_BASE+TVE_13C_REG_OFFSET, 0x004 );
   tve0_data.ptr_00c = ioremap( TVE0_BASE+TVE_00C_REG_OFFSET, 0x004 );
   tve0_data.ptr_020 = ioremap( TVE0_BASE+TVE_020_REG_OFFSET, 0x004 );
   tve0_data.ptr_10c = ioremap( TVE0_BASE+TVE_10C_REG_OFFSET, 0x004 );
   tve0_data.ptr_128 = ioremap( TVE0_BASE+TVE_128_REG_OFFSET, 0x004 );
   tve0_data.ptr_118 = ioremap( TVE0_BASE+TVE_118_REG_OFFSET, 0x004 );
   tve0_data.ptr_12c = ioremap( TVE0_BASE+TVE_12C_REG_OFFSET, 0x004 );
   tve0_data.ptr_008 = ioremap( TVE0_BASE+TVE_008_REG_OFFSET, 0x004 );
   tve0_data.ptr_024 = ioremap( TVE0_BASE+TVE_024_REG_OFFSET, 0x004 );
   tve0_data.ptr_110 = ioremap( TVE0_BASE+TVE_110_REG_OFFSET, 0x004 );
   tve0_data.ptr_120 = ioremap( TVE0_BASE+TVE_120_REG_OFFSET, 0x004 );

   tve1_data.ptr_004 = ioremap( TVE1_BASE+TVE_004_REG_OFFSET, 0x004 );
   tve1_data.ptr_014 = ioremap( TVE1_BASE+TVE_014_REG_OFFSET, 0x004 );
   tve1_data.ptr_01c = ioremap( TVE1_BASE+TVE_01C_REG_OFFSET, 0x004 );
   tve1_data.ptr_114 = ioremap( TVE1_BASE+TVE_114_REG_OFFSET, 0x004 );
   tve1_data.ptr_124 = ioremap( TVE1_BASE+TVE_124_REG_OFFSET, 0x004 );
   tve1_data.ptr_130 = ioremap( TVE1_BASE+TVE_130_REG_OFFSET, 0x004 );
   tve1_data.ptr_13c = ioremap( TVE1_BASE+TVE_13C_REG_OFFSET, 0x004 );
   tve1_data.ptr_00c = ioremap( TVE1_BASE+TVE_00C_REG_OFFSET, 0x004 );
   tve1_data.ptr_020 = ioremap( TVE1_BASE+TVE_020_REG_OFFSET, 0x004 );
   tve1_data.ptr_10c = ioremap( TVE1_BASE+TVE_10C_REG_OFFSET, 0x004 );
   tve1_data.ptr_128 = ioremap( TVE1_BASE+TVE_128_REG_OFFSET, 0x004 );
   tve1_data.ptr_118 = ioremap( TVE1_BASE+TVE_118_REG_OFFSET, 0x004 );
   tve1_data.ptr_12c = ioremap( TVE1_BASE+TVE_12C_REG_OFFSET, 0x004 );
   tve1_data.ptr_008 = ioremap( TVE1_BASE+TVE_008_REG_OFFSET, 0x004 );
   tve1_data.ptr_024 = ioremap( TVE1_BASE+TVE_024_REG_OFFSET, 0x004 );
   tve1_data.ptr_110 = ioremap( TVE1_BASE+TVE_110_REG_OFFSET, 0x004 );
   tve1_data.ptr_120 = ioremap( TVE1_BASE+TVE_120_REG_OFFSET, 0x004 );


   if ( (err = class_register(&cvbsctl_class)) ) { return(err); }


   tve0 = device_create( &cvbsctl_class, NULL,
                         MKDEV(0, 0), &tve0_data, "tve0" );
   if ( (err = sysfs_create_group(&tve0->kobj, &tve_attr_group)) ) { return(err); }

   tve1 = device_create( &cvbsctl_class, NULL,
                         MKDEV(0, 0), &tve1_data, "tve1" );
   if ( (err = sysfs_create_group(&tve1->kobj, &tve_attr_group)) ) { return(err); }


   tve_data_init( tve0 );
   tve_data_init( tve1 );


   return(0);
}

static void __exit sunxi_cvbsctl_exit(void)
{
   printk( KERN_INFO "exit cvbsctl\n" );


   device_destroy(&cvbsctl_class, tve1->devt);
   device_destroy(&cvbsctl_class, tve0->devt);


   class_unregister(&cvbsctl_class);


   iounmap(tve1_data.ptr_004);
   iounmap(tve1_data.ptr_014);
   iounmap(tve1_data.ptr_01c);
   iounmap(tve1_data.ptr_114);
   iounmap(tve1_data.ptr_124);
   iounmap(tve1_data.ptr_130);
   iounmap(tve1_data.ptr_13c);
   iounmap(tve1_data.ptr_00c);
   iounmap(tve1_data.ptr_020);
   iounmap(tve1_data.ptr_10c);
   iounmap(tve1_data.ptr_128);
   iounmap(tve1_data.ptr_118);
   iounmap(tve1_data.ptr_12c);
   iounmap(tve1_data.ptr_008);
   iounmap(tve1_data.ptr_024);
   iounmap(tve1_data.ptr_110);
   iounmap(tve1_data.ptr_120);

   iounmap(tve0_data.ptr_004);
   iounmap(tve0_data.ptr_014);
   iounmap(tve0_data.ptr_01c);
   iounmap(tve0_data.ptr_114);
   iounmap(tve0_data.ptr_124);
   iounmap(tve0_data.ptr_130);
   iounmap(tve0_data.ptr_13c);
   iounmap(tve0_data.ptr_00c);
   iounmap(tve0_data.ptr_020);
   iounmap(tve0_data.ptr_10c);
   iounmap(tve0_data.ptr_128);
   iounmap(tve0_data.ptr_118);
   iounmap(tve0_data.ptr_12c);
   iounmap(tve0_data.ptr_008);
   iounmap(tve0_data.ptr_024);
   iounmap(tve0_data.ptr_110);
   iounmap(tve0_data.ptr_120);

}

module_init(sunxi_cvbsctl_init);
module_exit(sunxi_cvbsctl_exit);


 
MODULE_LICENSE("GPL"); 
MODULE_AUTHOR("idr");

